from octue import exceptions
from octue.mixins import Labelable
from octue.resources.label import Label, LabelSet
from ..base import BaseTestCase


class MyLabelable(Labelable):
    pass


class LabelableTestCase(BaseTestCase):
    def test_instantiates(self):
        """Ensures the class instantiates without arguments"""
        Labelable()

    def test_instantiates_with_labels(self):
        """Ensures datafile inherits correctly from the Labelable class and passes arguments through"""
        labelable = MyLabelable(labels="")
        self.assertEqual(len(labelable.labels), 0)

        labelable = MyLabelable(labels=None)
        self.assertEqual(len(labelable.labels), 0)

        labelable = MyLabelable(labels="a b c")
        self.assertEqual(labelable.labels, {Label("a"), Label("b"), Label("c")})

    def test_instantiates_with_label_set(self):
        """Ensures datafile inherits correctly from the Labelable class and passes arguments through"""
        labelable_1 = MyLabelable(labels="")
        self.assertIsInstance(labelable_1.labels, LabelSet)
        labelable_2 = MyLabelable(labels=labelable_1.labels)
        self.assertFalse(labelable_1 is labelable_2)

    def test_fails_to_instantiates_with_non_iterable(self):
        """Ensures datafile inherits correctly from the Labelable class and passes arguments through"""

        class NoIter:
            pass

        with self.assertRaises(exceptions.InvalidLabelException) as error:
            MyLabelable(labels=NoIter())

        self.assertIn(
            "Labels must be expressed as a whitespace-delimited string or an iterable of strings",
            error.exception.args[0],
        )

    def test_reset_labels(self):
        """Ensures datafile inherits correctly from the Labelable class and passes arguments through"""
        labelable = MyLabelable(labels="a b")
        labelable.labels = "b c"
        self.assertEqual(labelable.labels, {Label("b"), Label("c")})

    def test_valid_labels(self):
        """Ensures valid labels do not raise an error"""
        labelable = MyLabelable()
        labelable.add_labels("a-valid-label")
        labelable.add_labels("label")
        labelable.add_labels("a1829tag")
        labelable.add_labels("1829")
        self.assertEqual(
            labelable.labels,
            {
                Label("a-valid-label"),
                Label("label"),
                Label("a1829tag"),
                Label("1829"),
            },
        )

    def test_mixture_valid_invalid(self):
        """Ensures that adding a variety of labels, some of which are invalid, doesn't partially add them to the object"""
        labelable = MyLabelable()
        labelable.add_labels("first-valid-should-be-added")
        try:
            labelable.add_labels("second-valid-should-not-be-added-because", "-the-third-is-invalid:")

        except exceptions.InvalidLabelException:
            pass

        self.assertEqual({Label("first-valid-should-be-added")}, labelable.labels)



from octue.mixins import MixinBase
from ..base import BaseTestCase


class MixinBaseTestCase(BaseTestCase):
    def test_instantiates_with_no_args(self):
        """Ensures the class instantiates without arguments"""
        MixinBase()

    def test_raises_exception_if_passed_args(self):
        """Ensures that the base mixin won't silently fail if passed arguments it's not supposed to have"""
        with self.assertRaises(TypeError):
            MixinBase("an_argument")

        with self.assertRaises(TypeError):
            MixinBase(an_argument="an_argument")



import collections.abc
import numbers

from octue import exceptions
from octue.utils.objects import get_nested_attribute, has_nested_attribute


def generate_complementary_filters(name, func):
    """Use a filter to generate its complementary filter, then return them together mapped to their names in a
    dictionary. The complementary filter is named f"not_{name}" or, if the name is "is", "is_not".

    :param str name:
    :param callable func:
    :return dict:
    """
    filter_action = {name: func}

    if name == "is":
        not_filter_name = "is_not"
    else:
        not_filter_name = f"not_{name}"

    not_filter_action = {
        not_filter_name: lambda item, value: not action(item, value) for name, action in filter_action.items()
    }

    return {**filter_action, **not_filter_action}


IS_FILTER_ACTIONS = generate_complementary_filters("is", lambda item, value: item is value)
EQUALS_FILTER_ACTIONS = generate_complementary_filters("equals", lambda item, value: value == item)
CONTAINS_FILTER_ACTIONS = generate_complementary_filters("contains", lambda item, value: value in item)
IN_RANGE_FILTER_ACTIONS = generate_complementary_filters("in_range", lambda item, value: value[0] <= item <= value[1])

ICONTAINS_FILTER_ACTIONS = generate_complementary_filters(
    "icontains", lambda item, value: value.casefold() in item.casefold()
)

COMPARISON_FILTER_ACTIONS = {
    "lt": lambda item, value: item < value,
    "lte": lambda item, value: item <= value,
    "gt": lambda item, value: item > value,
    "gte": lambda item, value: item >= value,
}


# Filters for specific types e.g. list or int.
TYPE_FILTERS = {
    "bool": {**EQUALS_FILTER_ACTIONS, **IS_FILTER_ACTIONS},
    "str": {
        **generate_complementary_filters("iequals", lambda item, value: value.casefold() == item.casefold()),
        **generate_complementary_filters("starts_with", lambda item, value: item.startswith(value)),
        **generate_complementary_filters("ends_with", lambda item, value: item.endswith(value)),
        **EQUALS_FILTER_ACTIONS,
        **IS_FILTER_ACTIONS,
        **COMPARISON_FILTER_ACTIONS,
        **CONTAINS_FILTER_ACTIONS,
        **ICONTAINS_FILTER_ACTIONS,
        **IN_RANGE_FILTER_ACTIONS,
    },
    "NoneType": {**EQUALS_FILTER_ACTIONS, **IS_FILTER_ACTIONS},
    "LabelSet": {
        **EQUALS_FILTER_ACTIONS,
        **CONTAINS_FILTER_ACTIONS,
        **IS_FILTER_ACTIONS,
        **generate_complementary_filters("any_label_contains", lambda item, value: item.any_label_contains(value)),
        **generate_complementary_filters(
            "any_label_starts_with", lambda item, value: item.any_label_starts_with(value)
        ),
        **generate_complementary_filters("any_label_ends_with", lambda item, value: item.any_label_ends_with(value)),
    },
    "datetime": {
        **EQUALS_FILTER_ACTIONS,
        **IS_FILTER_ACTIONS,
        **COMPARISON_FILTER_ACTIONS,
        **IN_RANGE_FILTER_ACTIONS,
        "year_equals": lambda item, value: item.year == value,
        "year_in": lambda item, value: item.year in value,
        "month_equals": lambda item, value: item.month == value,
        "month_in": lambda item, value: item.month in value,
        "day_equals": lambda item, value: item.day == value,
        "day_in": lambda item, value: item.day in value,
        "weekday_equals": lambda item, value: item.weekday() == value,
        "weekday_in": lambda item, value: item.weekday() in value,
        "iso_weekday_equals": lambda item, value: item.isoweekday() == value,
        "iso_weekday_in": lambda item, value: item.isoweekday() in value,
        "time_equals": lambda item, value: item.time() == value,
        "time_in": lambda item, value: item.time() in value,
        "hour_equals": lambda item, value: item.hour == value,
        "hour_in": lambda item, value: item.hour in value,
        "minute_equals": lambda item, value: item.minute == value,
        "minute_in": lambda item, value: item.minute in value,
        "second_equals": lambda item, value: item.second == value,
        "second_in": lambda item, value: item.second in value,
        "in_date_range": lambda item, value: value[0] <= item.date() <= value[1],
        "in_time_range": lambda item, value: value[0] <= item.time() <= value[1],
    },
}

# Filters for interfaces e.g. iterables or numbers.
INTERFACE_FILTERS = {
    numbers.Number: {
        **EQUALS_FILTER_ACTIONS,
        **COMPARISON_FILTER_ACTIONS,
        **IS_FILTER_ACTIONS,
        **IN_RANGE_FILTER_ACTIONS,
    },
    collections.abc.Iterable: {
        **EQUALS_FILTER_ACTIONS,
        **CONTAINS_FILTER_ACTIONS,
        **ICONTAINS_FILTER_ACTIONS,
        **IS_FILTER_ACTIONS,
    },
}


class Filterable:
    def satisfies(self, raise_error_if_filter_is_invalid=True, **kwargs):
        """Check that the instance satisfies the given filter for the given filter value. The filter should be provided
        as a single keyword argument such as `name__first__equals="Joe"`

        :param bool raise_error_if_filter_is_invalid:
        :param {str: any} kwargs: a single keyword argument whose key is the name of the filter and whose value is the value to filter for
        :raise ValueError: if more than one keyword argument is received
        :raise AttributeError: if the instance doesn't have the attribute specified in the filter name (only if `raise_error_if_filter_is_invalid` is `True`)
        :raise octue.exceptions.InvalidInputException: if the filter name is invalid or there are no filters for the type of attribute specified in the filter name
        :return mixed:
        """
        if len(kwargs) != 1:
            raise ValueError(f"The satisfies method only takes one keyword argument; received {kwargs!r}.")

        filter_name, filter_value = list(kwargs.items())[0]

        try:
            attribute_name, filter_action = self._split_filter_name(filter_name)

            try:
                attribute = get_nested_attribute(self, attribute_name)

            except AttributeError as error:
                if raise_error_if_filter_is_invalid:
                    raise error

                return False

            filter_ = self._get_filter(attribute, filter_action)
            return filter_(attribute, filter_value)

        except exceptions.InvalidInputException as error:
            return self._try_equals_filter_shortcut(filter_name, filter_value, error)

    def _split_filter_name(self, filter_name):
        """Split the filter name into the attribute name and filter action, raising an error if it the attribute name
        and filter action aren't delimited by a double underscore i.e. "__".
        """
        *attribute_names, filter_action = filter_name.split("__")

        if not attribute_names:

            raise exceptions.InvalidInputException(
                f"Invalid filter name {filter_name!r}. Filter names should be in the form "
                f"'<attribute_name_0>__<attribute_name_1>__<...>__<filter_kind>' with at least one attribute name "
                f"included."
            )

        return ".".join(attribute_names), filter_action

    def _get_filter(self, attribute, filter_action):
        """Get the filter for the attribute and filter action, raising an error if there is no filter action of that
        name.
        """
        try:
            return self._get_filter_actions_for_attribute(attribute)[filter_action]

        except KeyError as error:
            raise exceptions.InvalidInputException(
                f"There is no filter called {error.args[0]!r} for attributes of type {type(attribute)}. The options "
                f"are {self._get_filter_actions_for_attribute(attribute).keys()!r}"
            )

    def _get_filter_actions_for_attribute(self, attribute):
        """Get the possible filters for the given attribute based on its type or interface, raising an error if the
        attribute's type isn't supported (i.e. if there aren't any filters defined for it)."""
        try:
            return TYPE_FILTERS[type(attribute).__name__]

        except KeyError as error:
            # This allows handling of objects that conform to a certain interface (e.g. iterables) without needing the
            # specific type.
            for type_ in INTERFACE_FILTERS:
                if not isinstance(attribute, type_):
                    continue
                return INTERFACE_FILTERS[type_]

            raise exceptions.InvalidInputException(
                f"Attributes of type {error.args[0]} are not currently supported for filtering."
            )

    def _try_equals_filter_shortcut(self, filter_name, filter_value, error):
        """Try to use the equals filter shortcut e.g. `a=7` instead of `a__equals=7` or `a__b=7` instead of
        `a__b__equals=7`. Raise the error if this is not applicable (e.g. the filter name is just wrong).

        :param str filter_name:
        :param mixed filter_value:
        :param Exception error:  the error to raise if the equals filter shortcut is not applicable
        :raise Exception: if the equals filter shortcut is not applicable
        :return mixed:
        """
        possible_attribute_name = ".".join(filter_name.split("__"))

        if has_nested_attribute(self, possible_attribute_name):
            attribute = get_nested_attribute(self, possible_attribute_name)
            filter_ = self._get_filter(attribute, "equals")
            return filter_(attribute, filter_value)

        raise error



from octue.cloud.pub_sub.service import Service
from octue.resources import service_backends


BACKEND_TO_SERVICE_MAPPING = {"GCPPubSubBackend": Service}


class Child:
    """A class representing a child service that can be asked questions. It is a convenience wrapper for `Service` that
    makes the asking of questions more intuitive for Scientists and allows easier selection of backends.

    :param str name:
    :param str id:
    :param dict backend: must include the key "name" with a value of the name of the type of backend e.g. GCPPubSubBackend and key-value pairs for any other parameters the chosen backend expects
    :return None:
    """

    def __init__(self, name, id, backend):
        self.name = name
        self.id = id

        backend_type_name = backend.pop("name")
        backend = service_backends.get_backend(backend_type_name)(**backend)
        self._service = BACKEND_TO_SERVICE_MAPPING[backend_type_name](backend=backend)

    def ask(self, input_values=None, input_manifest=None, subscribe_to_logs=True, timeout=20):
        """Ask the child a question (i.e. send it some input value and/or a manifest and wait for it to run an analysis
        on them and return the output values). The input values given must adhere to the Twine file of the child.

        :param any input_values: the input values of the question
        :param octue.resources.manifest.Manifest|None input_manifest: the input manifest of the question
        :param bool subscribe_to_logs: if `True`, subscribe to logs from the remote service and handle them with the local log handlers
        :param float timeout: time in seconds to wait for an answer before raising a timeout error
        :raise TimeoutError: if the timeout is exceeded while waiting for an answer
        :return dict: dictionary containing the keys "output_values" and "output_manifest"
        """
        subscription, _ = self._service.ask(self.id, input_values, input_manifest, subscribe_to_logs)
        return self._service.wait_for_answer(subscription, timeout)



import os
import subprocess
import time
import unittest
from unittest.mock import Mock, call

from octue.utils.processes import run_subprocess_and_log_stdout_and_stderr
from tests.base import BaseTestCase


@unittest.skipIf(condition=os.name == "nt", reason="See issue https://github.com/octue/octue-sdk-python/issues/229")
class TestRunSubprocessAndLogStdoutAndStderr(BaseTestCase):
    def test_error_raised_if_process_fails(self):
        """Test that an error is raised if the process fails."""
        mock_logger = Mock()

        with self.assertRaises(subprocess.CalledProcessError):
            run_subprocess_and_log_stdout_and_stderr(command=["blah blah blah"], logger=mock_logger, shell=True)

        log_message = mock_logger.info.call_args.args[0]
        self.assertTrue("not found" in log_message)

    def test_stdout_is_logged(self):
        """Test that any output to stdout is logged."""
        mock_logger = Mock()

        process = run_subprocess_and_log_stdout_and_stderr(
            command=["echo hello && echo goodbye"], logger=mock_logger, shell=True
        )

        self.assertEqual(process.returncode, 0)
        time.sleep(1)
        mock_logger.info.assert_has_calls([call("hello"), call("goodbye")])



import datetime
import functools
import logging
import os
import tempfile
from urllib.parse import urlparse
import google.api_core.exceptions
import pkg_resources
from google_crc32c import Checksum

from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.cloud.storage.path import CLOUD_STORAGE_PROTOCOL
from octue.exceptions import CloudLocationNotSpecified, FileNotFoundException, InvalidInputException
from octue.mixins import Filterable, Hashable, Identifiable, Labelable, Loggable, Pathable, Serialisable, Taggable
from octue.mixins.hashable import EMPTY_STRING_HASH_VALUE
from octue.utils import isfile


module_logger = logging.getLogger(__name__)


OCTUE_METADATA_NAMESPACE = "octue"

ID_DEFAULT = None
TAGS_DEFAULT = None
LABELS_DEFAULT = None


class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiable, Hashable, Filterable):
    """A representation of a data file on the Octue system. If the given path is a cloud path and `hypothetical` is not
    `True`, the datafile's metadata is pulled from the given cloud location, and any conflicting parameters (see the
    `Datafile.metadata` method description for the parameter names concerned) are ignored. The metadata of cloud
    datafiles can be changed using the `Datafile.update_metadata` method, but not during instantiation.

    Files in a manifest look like this:

        {
          "path": "folder/subfolder/file_1.csv",
          "extension": "csv",
          "tags": {},
          "labels": [],
          "timestamp": datetime.datetime(2021, 5, 3, 18, 15, 58, 298086),
          "id": "abff07bc-7c19-4ed5-be6d-a6546eae8e86",
          "size_bytes": 59684813,
          "sha-512/256": "somesha"
        },

    :param str|None path: The path of this file locally or in the cloud, which may include folders or subfolders, within the dataset. If no path_from parameter is set, then absolute paths are acceptable, otherwise relative paths are required.
    :param str|None local_path: If a cloud path is given as the `path` parameter, this is the path to an existing local file that is known to be in sync with the cloud object
    :param str|None cloud_path: If a local path is given for the `path` parameter, this is a cloud path to keep in sync with the local file
    :param str|None project_name: The name of the project if the datafile also exists in the cloud
    :param datetime.datetime|int|float|None timestamp: A posix timestamp associated with the file, in seconds since epoch, typically when it was created but could relate to a relevant time point for the data
    :param str id: The Universally Unique ID of this file (checked to be valid if not None, generated if None)
    :param logging.Logger logger: A logger instance to which operations with this datafile will be logged. Defaults to the module logger.
    :param Pathable path_from: The root Pathable object (typically a Dataset) that this Datafile's path is relative to.
    :param str|None project_name: The name of the cloud project if the datafile is located in the cloud
    :param dict|TagDict tags: key-value pairs with string keys conforming to the Octue tag format (see TagDict)
    :param iter(str) labels: Space-separated string of labels relevant to this file
    :param bool skip_checks:
    :param str mode: if using as a context manager, open the datafile for reading/editing in this mode (the mode options are the same as for the builtin open function)
    :param bool update_cloud_metadata: if using as a context manager and this is True, update the cloud metadata of the datafile when the context is exited
    :param bool hypothetical: True if the file does not actually exist or access is not available at instantiation
    :return None:
    """

    _SERIALISE_FIELDS = (
        "id",
        "name",
        "path",
        "cloud_path",
        "project_name",
        "tags",
        "labels",
        "timestamp",
        "_cloud_metadata",
    )

    def __init__(
        self,
        path,
        local_path=None,
        cloud_path=None,
        project_name=None,
        timestamp=None,
        id=ID_DEFAULT,
        logger=None,
        path_from=None,
        tags=TAGS_DEFAULT,
        labels=LABELS_DEFAULT,
        skip_checks=True,
        mode="r",
        update_cloud_metadata=True,
        hypothetical=False,
        **kwargs,
    ):
        super().__init__(
            id=id,
            name=kwargs.pop("name", None),
            immutable_hash_value=kwargs.pop("immutable_hash_value", None),
            logger=logger,
            tags=tags,
            labels=labels,
            path=path,
            path_from=path_from,
        )

        self.timestamp = timestamp
        self.extension = os.path.splitext(path)[-1].strip(".")
        self.project_name = project_name

        self._local_path = None
        self._cloud_path = None
        self._hypothetical = hypothetical
        self._open_attributes = {"mode": mode, "update_cloud_metadata": update_cloud_metadata, **kwargs}
        self._cloud_metadata = {}

        if self.path.startswith(CLOUD_STORAGE_PROTOCOL):
            if project_name is None:
                raise CloudLocationNotSpecified(
                    f"The `project_name` parameter is required to instantiate a Datafile from a cloud object; received "
                    f"{project_name}."
                )

            self._store_cloud_location(project_name=project_name, cloud_path=path)

            if not self._hypothetical:
                # Collect any non-`None` metadata instantiation parameters so the user can be warned if they conflict
                # with any metadata already on the cloud object.
                initialisation_parameters = {}

                for parameter in ("id", "timestamp", "tags", "labels"):
                    value = locals().get(parameter)
                    if value is not None:
                        initialisation_parameters[parameter] = value

                self._use_cloud_metadata(**initialisation_parameters)

            if local_path:
                # If there is no file at the given local path or the file is different to the one in the cloud, download
                # the cloud file locally.
                if not os.path.exists(local_path) or self._cloud_metadata.get("crc32c") != calculate_hash(local_path):
                    self.download(local_path)
                else:
                    self._local_path = local_path

        else:
            self._local_path = self.absolute_path

            # Run integrity checks on the file.
            if not skip_checks:
                self.check(**kwargs)

            if cloud_path:
                self.cloud_path = cloud_path

    def __enter__(self):
        self._open_context_manager = self.open(**self._open_attributes)
        return self, self._open_context_manager.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._open_context_manager.__exit__(exc_type, exc_val, exc_tb)

    def __lt__(self, other):
        if not isinstance(other, Datafile):
            raise TypeError(f"An object of type {type(self)} cannot be compared with {type(other)}.")
        return self.absolute_path < other.absolute_path

    def __gt__(self, other):
        if not isinstance(other, Datafile):
            raise TypeError(f"An object of type {type(self)} cannot be compared with {type(other)}.")
        return self.absolute_path > other.absolute_path

    def __repr__(self):
        return f"<{type(self).__name__}({self.name!r})>"

    @classmethod
    def deserialise(cls, serialised_datafile, path_from=None):
        """Deserialise a Datafile from a dictionary. The `path_from` parameter is only used if the path in the
        serialised Dataset is relative.

        :param dict serialised_datafile:
        :param octue.mixins.Pathable path_from:
        :return Datafile:
        """
        cloud_metadata = serialised_datafile.pop("_cloud_metadata", {})

        if not os.path.isabs(serialised_datafile["path"]) and not serialised_datafile["path"].startswith(
            CLOUD_STORAGE_PROTOCOL
        ):
            datafile = Datafile(**serialised_datafile, path_from=path_from)
        else:
            datafile = Datafile(**serialised_datafile)

        datafile._cloud_metadata = cloud_metadata
        return datafile

    def to_cloud(
        self, project_name=None, cloud_path=None, bucket_name=None, path_in_bucket=None, update_cloud_metadata=True
    ):
        """Upload a datafile to Google Cloud Storage. Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must
        be provided.

        :param str|None project_name: name of Google Cloud project to store datafile in
        :param str|None cloud_path: full path to cloud storage location to store datafile at (e.g. `gs://bucket_name/path/to/file.csv`)
        :param str|None bucket_name: name of bucket to store datafile in
        :param str|None path_in_bucket: cloud storage path to store datafile at (e.g. `path/to/file.csv`)
        :param bool update_cloud_metadata: if `True`, update the metadata of the datafile in the cloud at upload time
        :return str: gs:// path for datafile
        """
        project_name, cloud_path = self._get_cloud_location(project_name, cloud_path, bucket_name, path_in_bucket)
        self.get_cloud_metadata()

        # If the datafile's file has been changed locally, overwrite its cloud copy.
        if self._cloud_metadata.get("crc32c") != self.hash_value:
            GoogleCloudStorageClient(project_name=project_name).upload_file(
                local_path=self.local_path,
                cloud_path=cloud_path,
                metadata=self.metadata(),
            )

        if update_cloud_metadata:
            # If the datafile's metadata has been changed locally, update the cloud file's metadata.
            local_metadata = self.metadata()

            if self._cloud_metadata.get("custom_metadata") != local_metadata:
                self.update_cloud_metadata()

        return self.cloud_path

    def get_cloud_metadata(self):
        """Get the cloud metadata for the datafile.

        :return dict:
        """
        if not self.cloud_path:
            self._raise_cloud_location_error()

        cloud_metadata = GoogleCloudStorageClient(self.project_name).get_metadata(cloud_path=self.cloud_path)

        if cloud_metadata:
            self._cloud_metadata = cloud_metadata

    def update_cloud_metadata(self):
        """Update the cloud metadata for the datafile.

        :return None:
        """
        if not self.cloud_path:
            self._raise_cloud_location_error()

        GoogleCloudStorageClient(project_name=self.project_name).overwrite_custom_metadata(
            metadata=self.metadata(), cloud_path=self.cloud_path
        )

    @property
    def name(self):
        return self._name or str(os.path.split(self.path)[-1])

    @property
    def cloud_path(self):
        """Get the cloud path of the datafile.

        :return str|None:
        """
        return self._cloud_path

    @cloud_path.setter
    def cloud_path(self, path):
        """Set the cloud path of the datafile.

        :param str|None path:
        :return None:
        """
        if path is None:

            if not self.exists_locally:
                raise CloudLocationNotSpecified(
                    "The cloud path cannot be reset because this datafile only exists in the cloud."
                )

            self._cloud_path = None

        else:
            self.to_cloud(cloud_path=path)

    @property
    def cloud_protocol(self):
        """Get the cloud protocol of the datafile if it exists in the cloud (e.g. "gs" for a cloud path of
        "gs://my-bucket/my-file.txt").

        :return str|None:
        """
        if not self.exists_in_cloud:
            return None
        return urlparse(self.cloud_path).scheme

    @property
    def bucket_name(self):
        """Get the name of the bucket the datafile exists in if it exists in the cloud.

        :return str|None:
        """
        if self.cloud_path:
            return storage.path.split_bucket_name_from_gs_path(self.cloud_path)[0]
        return None

    @property
    def path_in_bucket(self):
        """Get the path of the datafile in its bucket if it exists in the cloud.

        :return str|None:
        """
        if self.cloud_path:
            return storage.path.split_bucket_name_from_gs_path(self.cloud_path)[1]
        return None

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        """Set the datafile's timestamp.

        :param datetime.datetime|int|float|None value:
        :raise TypeError: if value is of an incorrect type
        :return None:
        """
        if isinstance(value, datetime.datetime) or value is None:
            self._timestamp = value
        elif isinstance(value, (int, float)):
            self._timestamp = datetime.datetime.fromtimestamp(value)
        else:
            raise TypeError(
                f"timestamp should be a datetime.datetime instance, an int, a float, or None; received {value!r}"
            )

    @property
    def posix_timestamp(self):
        """Get the timestamp of the datafile in posix format.

        :return float:
        """
        if self.timestamp is None:
            return None

        return self.timestamp.timestamp()

    @property
    def _last_modified(self):
        """Get the date/time the file was last modified in units of seconds since epoch (posix time).

        :return float:
        """
        if self._path_is_in_google_cloud_storage:
            last_modified = self._cloud_metadata.get("updated")

            if last_modified is None:
                return None

            return last_modified.timestamp()

        return os.path.getmtime(self.absolute_path)

    @property
    def size_bytes(self):
        if self._path_is_in_google_cloud_storage:
            return self._cloud_metadata.get("size")

        return os.path.getsize(self.absolute_path)

    @property
    def exists_in_cloud(self):
        """Does the file exist in the cloud?

        :return bool:
        """
        return self.cloud_path is not None

    @property
    def exists_locally(self):
        """Does the file exist locally?

        :return bool:
        """
        return self._local_path is not None

    def download(self, local_path=None):
        """Download the file from the cloud to the given local path or a random temporary path.

        :param str|None local_path:
        :raise CloudLocationNotSpecified: if the datafile does not exist in the cloud
        :return str: path to local file
        """
        if not self.exists_in_cloud:
            raise CloudLocationNotSpecified("Cannot download a file that doesn't exist in the cloud.")

        # Avoid downloading to a local path if the datafile has already been downloaded to it.
        if (local_path is None and self._local_path is not None) or (
            local_path is not None and local_path == self._local_path
        ):
            return self._local_path

        if local_path is not None:
            self._local_path = os.path.abspath(local_path)
        else:
            self._local_path = tempfile.NamedTemporaryFile(delete=False).name

        try:
            GoogleCloudStorageClient(project_name=self.project_name).download_to_file(
                local_path=self._local_path, cloud_path=self.cloud_path
            )

        except google.api_core.exceptions.NotFound as e:
            # If in reading mode, raise an error if no file exists at the path; if in a writing mode, create a new file.
            if self._open_attributes["mode"] == "r":
                raise e

        # Now use hash value of local file instead of cloud file.
        self.reset_hash()
        return self._local_path

    @property
    def local_path(self):
        """Get the local path for the datafile, downloading it from the cloud to a temporary file if necessary. If
        downloaded, the local path is added to a cache to avoid downloading again in the same runtime.
        """
        if self._local_path:
            return self._local_path

        return self.download()

    @local_path.setter
    def local_path(self, path):
        """Set the local path of the datafile to an empty path (a path not corresponding to an existing file) and
        download the contents of the corresponding cloud file to the local path.

        :param str path:
        :raise FileExistsError: if the path corresponds to an existing file
        :return None:
        """
        if path is None:
            if not self.exists_in_cloud:
                raise CloudLocationNotSpecified(
                    "The local path cannot be reset because this datafile only exists locally."
                )

            self._local_path = None
            return

        if os.path.exists(path):
            raise FileExistsError(
                "Only a path not corresponding to an existing file can be used. This is because the contents of the "
                "existing cloud file will be downloaded to the new local path and would overwrite any existing file at "
                "the given path."
            )

        GoogleCloudStorageClient(project_name=self.project_name).download_to_file(
            local_path=path, cloud_path=self.cloud_path
        )

        self._local_path = os.path.abspath(path)

    def _use_cloud_metadata(self, **initialisation_parameters):
        """Populate the datafile's attributes from the metadata of the cloud object located at its path (by necessity a
        cloud path) and project name. If there is a conflict between the cloud metadata and a given local initialisation
        parameter, the local value is used.

        :param initialisation_parameters: key-value pairs of initialisation parameter names and values (provide to check for conflicts with cloud metadata)
        :return None:
        """
        self.get_cloud_metadata()
        cloud_custom_metadata = self._cloud_metadata.get("custom_metadata", {})
        self._warn_about_attribute_conflicts(cloud_custom_metadata, **initialisation_parameters)

        self._set_id(
            initialisation_parameters.get(
                "id", cloud_custom_metadata.get(f"{OCTUE_METADATA_NAMESPACE}__id", ID_DEFAULT)
            )
        )

        self.immutable_hash_value = self._cloud_metadata.get("crc32c", EMPTY_STRING_HASH_VALUE)

        for attribute in ("timestamp", "tags", "labels"):
            setattr(
                self,
                attribute,
                initialisation_parameters.get(
                    attribute, cloud_custom_metadata.get(f"{OCTUE_METADATA_NAMESPACE}__{attribute}")
                ),
            )

    def _warn_about_attribute_conflicts(self, cloud_custom_metadata, **initialisation_parameters):
        """Raise a warning if there is a conflict between the cloud custom metadata and the given initialisation
        parameters if the cloud value is not `None` or an empty collection.

        :param dict cloud_custom_metadata:
        :return None:
        """
        for attribute_name, attribute_value in initialisation_parameters.items():

            cloud_metadata_value = cloud_custom_metadata.get(f"{OCTUE_METADATA_NAMESPACE}__{attribute_name}")

            if cloud_metadata_value == attribute_value or not cloud_metadata_value:
                continue

            module_logger.warning(
                f"The value {cloud_metadata_value!r} of the {type(self).__name__} attribute {attribute_name!r} from "
                f"the cloud conflicts with the value given locally at instantiation {attribute_value!r}. The local "
                f"value has been used and will overwrite the cloud value if the datafile is saved."
            )

    def _calculate_hash(self):
        """Calculate the hash of the file."""
        hash = calculate_hash(self.local_path)
        return super()._calculate_hash(hash)

    def _get_cloud_location(self, project_name=None, cloud_path=None, bucket_name=None, path_in_bucket=None):
        """Get the cloud location details for the bucket, allowing the keyword arguments to override any stored values.
        Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must be provided. Once the cloud location details
        have been determined, update the stored cloud location details.

        :param str|None project_name:
        :param str|None cloud_path:
        :param str|None bucket_name:
        :param str|None path_in_bucket:
        :raise octue.exceptions.CloudLocationNotSpecified: if an exact cloud location isn't provided and isn't available implicitly (i.e. the Datafile wasn't loaded from the cloud previously)
        :return (str, str): project name and cloud path
        """
        project_name = project_name or self.project_name
        cloud_path = cloud_path or self.cloud_path

        if not cloud_path:
            try:
                cloud_path = storage.path.generate_gs_path(bucket_name, path_in_bucket)
            except (TypeError, AttributeError):
                self._raise_cloud_location_error()

        self._store_cloud_location(project_name=project_name, cloud_path=cloud_path)
        return project_name, cloud_path

    def _store_cloud_location(self, project_name=None, cloud_path=None, bucket_name=None, path_in_bucket=None):
        """Store the cloud location of the datafile. Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must be
        provided.

        :param str|None project_name:
        :param str|None cloud_path:
        :param str|None bucket_name:
        :param str|None path_in_bucket:
        :return None:
        """
        self.project_name = project_name

        if cloud_path:
            self._cloud_path = cloud_path
        else:
            try:
                self._cloud_path = storage.path.generate_gs_path(bucket_name, path_in_bucket)
            except (TypeError, AttributeError):
                self._raise_cloud_location_error()

    def _raise_cloud_location_error(self):
        """Raise an error indicating that the cloud location of the datafile has not yet been specified.

        :raise CloudLocationNotSpecified:
        :return None:
        """
        raise CloudLocationNotSpecified(
            f"{self!r} wasn't previously loaded from the cloud so doesn't have an implicit cloud location - please "
            f"specify its exact location (its project name and cloud path)."
        )

    def check(self, size_bytes=None, sha=None, last_modified=None, extension=None):
        """Check file presence and integrity"""
        # TODO Check consistency of size_bytes input against self.size_bytes property for a file if we have one
        # TODO Check consistency of sha against file contents if we have a file
        # TODO Check consistency of last_modified date

        if (extension is not None) and not self.path.endswith(extension):
            raise InvalidInputException(
                f"Extension provided ({extension}) does not match file extension (from {self.path}). Pass extension="
                f"None to set extension from filename automatically."
            )

        if not self.exists():
            raise FileNotFoundException(f"No file found at {self.absolute_path}")

    def exists(self):
        """Returns true if the datafile exists on the current system, false otherwise"""
        return isfile(self.absolute_path)

    @property
    def open(self):
        """Get a context manager for handling the opening and closing of the datafile for reading/editing.

        :return type: the class octue.resources.datafile._DatafileContextManager
        """
        return functools.partial(_DatafileContextManager, self)

    def metadata(self, use_octue_namespace=True):
        """Get the datafile's metadata in a serialised form (i.e. the attributes `id`, `timestamp`, `labels`, `tags`,
        and `sdk_version`).

        :param bool use_octue_namespace: if True, prefix metadata names with "octue__"
        :return dict:
        """
        metadata = {
            "id": self.id,
            "timestamp": self.timestamp,
            "tags": self.tags,
            "labels": self.labels,
            "sdk_version": pkg_resources.get_distribution("octue").version,
        }

        if not use_octue_namespace:
            return metadata

        return {f"{OCTUE_METADATA_NAMESPACE}__{key}": value for key, value in metadata.items()}


class _DatafileContextManager:
    """A context manager for opening datafiles for reading and writing locally or from the cloud. Its usage is analogous
    to the builtin open context manager. If opening a local datafile in write mode, the manager will attempt to
    determine if the folder path exists and, if not, will create the folder structure required to write the file.

    Usage:
    ```
    my_datafile = Datafile(path='subfolder/subsubfolder/my_datafile.json)
    with my_datafile.open('w') as fp:
        fp.write("{}")
    ```

    This is equivalent to the standard python:
    ```
    my_datafile = Datafile(path='subfolder/subsubfolder/my_datafile.json)
    os.makedirs(os.path.split(my_datafile.absolute_path)[0], exist_ok=True)
    with open(my_datafile.absolute_path, 'w') as fp:
        fp.write("{}")
    ```

    :param octue.resources.datafile.Datafile datafile:
    :param str mode: open the datafile for reading/editing in this mode (the mode options are the same as for the builtin `open` function)
    :param bool update_cloud_metadata: if this is True, update the cloud metadata of the datafile when the context is exited
    :return None:
    """

    MODIFICATION_MODES = {"w", "a", "x", "+", "U"}

    def __init__(self, datafile, mode="r", update_cloud_metadata=True, **kwargs):
        self.datafile = datafile
        self.mode = mode
        self._update_cloud_metadata = update_cloud_metadata
        self.kwargs = kwargs
        self._fp = None

    def __enter__(self):
        """Open the datafile, first downloading it from the cloud if necessary.

        :return io.TextIOWrapper:
        """
        if "w" in self.mode:
            os.makedirs(os.path.split(self.datafile.local_path)[0], exist_ok=True)

        self._fp = open(self.datafile.local_path, self.mode, **self.kwargs)
        return self._fp

    def __exit__(self, *args):
        """Close the datafile, updating the corresponding file in the cloud if necessary and its metadata if
        self._update_cloud_metadata is True.

        :return None:
        """
        if self._fp is not None:
            self._fp.close()

        if self.datafile.exists_in_cloud and any(character in self.mode for character in self.MODIFICATION_MODES):
            self.datafile.to_cloud(update_cloud_metadata=self._update_cloud_metadata)


def calculate_hash(path):
    """Calculate the hash of the file at the given path."""
    hash = Checksum()

    with open(path, "rb") as f:
        # Read and update hash value in blocks of 4K.
        for byte_block in iter(lambda: f.read(4096), b""):
            hash.update(byte_block)

    return hash
